from parchments.core.validation import is_valid_list_choice, is_valid_date_or_datetime
from parchments.core.block import Block
from parchments.core.choices import VALUE_TYPE_CHOICES


class Row:
    block_order_list = list()
    period_iteration = None
    over_period_iteration = None

    def __init__(self, name, value_type, value_decimals, period_iteration, over_period_iteration):
        if is_valid_list_choice(value_type, VALUE_TYPE_CHOICES):
            self.value_type = value_type
        self.value_decimals = value_decimals
        self.name = name
        self.period_iteration = period_iteration
        self.over_period_iteration = over_period_iteration
        self.block_dict = dict()

    def create_block(self, period, value):
        self.block_dict[period.key] = Block(period, value, self.value_type, self.value_decimals)
        if period.key not in self.block_order_list:
            self.block_order_list.append(period.key)
        self.block_order_list.sort()

    def update_row(self):
        for loop_index, block_order in enumerate(self.block_order_list):
            if loop_index + 1 < len(self.block_order_list):
                self.block_dict[block_order].compare_projected(self.block_dict[self.block_order_list[loop_index + 1]])
            else:
                self.block_dict[block_order].compare_projected(self.block_dict[block_order])
            if self.over_period_iteration == 'year':
                if str(int(block_order) + 10000) in self.block_order_list:
                    self.block_dict[block_order].compare_over_projected(
                        self.block_dict[str(int(block_order) + 10000)])
                else:
                    self.block_dict[block_order].compare_over_projected(self.block_dict[block_order])

    def as_dict(self):
        row_list = list()
        for block_order in self.block_order_list:
            row_list.append(self.block_dict[block_order].as_dict())
        return row_list



VALUE_TYPE_CHOICES = [
    'string',
    'int',
    'dollar',
    'percentage',
    'bool',
]

PERIOD_ITERATION_CHOICES = [
    'year',
    'quarter',
    'month',
    'day'
]

OVER_PERIOD_ITERATION_CHOICES = [
    'year',
    'quarter',
    'month',
]



from datetime import datetime

# Todo: This will need to be remove as it's no longer in use.
def period_key(period: datetime, iteration: str):
    if iteration == 'year':
        return '%s0101' % period.strftime('%Y')
    if iteration == 'month':
        return '%s01' % period.strftime('%Y%m')
    if iteration == 'day':
        return '%s' % period.strftime('%Y%m%d')




from datetime import date, datetime


def is_valid_date_or_datetime(date):
    if type(date) is date or type(date) is datetime:
        return True
    else:
        raise ValueError('Invalid. Must be a date or datetime')


def is_valid_list_choice(variable, choice_list):
    if variable in choice_list:
        return True
    else:
        raise SyntaxError('Invalid option. Choices are %s' % choice_list)


class Value:

    def __init__(self, value, value_type, decimals=2):
        self.data = {
            'raw': value,
            'clean': value,
            'verbose': self.verbose(value, value_type),
            'type': value_type,
            'decimals': decimals,
        }

    def verbose(self, value, value_type):
        if value_type == 'dollar':
            return '${:,.2f}'.format(value)
        elif value_type == 'percentage':
            return '{:,.4f}%'.format((value * 100))
        elif value_type == 'int':
            return '{:,.0f}'.format(value)
        else:
            return value

    def as_dict(self):
        return self.data


from parchments.core.row import Row
from parchments.core.period import Period
from parchments.core.choices import PERIOD_ITERATION_CHOICES, OVER_PERIOD_ITERATION_CHOICES


class Grid:

    def __init__(self, row_index, period_iteration='month', over_period_iteration='year'):
        if period_iteration in PERIOD_ITERATION_CHOICES:
            self.period_iteration = period_iteration
        else:
            raise SyntaxError('Invalid period iteration choices %s' % PERIOD_ITERATION_CHOICES)

        if over_period_iteration in OVER_PERIOD_ITERATION_CHOICES:
            self.over_period_iteration = over_period_iteration
        else:
            raise SyntaxError('Invalid layer iteration choices %s' % OVER_PERIOD_ITERATION_CHOICES)

        self.row_index = row_index

        self.row_dict = dict()
        for row in self.row_index:
            self.row_dict[row[0]] = Row(row[0], row[1], row[2], self.period_iteration, self.over_period_iteration)

    def add_period(self, datetime, value_list):
        period = Period(datetime, self.period_iteration)
        if type(value_list) is list:
            for loop_index, row in enumerate(self.row_index):
                self.row_dict[row[0]].create_block(period, value_list[loop_index])
                self.row_dict[row[0]].update_row()

    def to_dict(self):
        grid_dict = dict()
        grid_dict['row_data'] = dict()
        for row in self.row_index:
            # Todo: This is a bit lazy for assignment of block order list
            grid_dict['row_data'][row[0]] = self.row_dict[row[0]].as_dict()
            grid_dict['column_index'] = self.row_dict[row[0]].block_order_list
        return grid_dict




import unittest
import parchments
import datetime
import decimal

test_index = (
    ('goats', 'int', 0),
    ('price', 'dollar', 2),
    ('value', 'percentage', 4),
    ('names', 'string', 0),
)

test_grid = parchments.Grid(test_index)


class TestGrid(unittest.TestCase):

    def test_grid_add_period(self, period=datetime.datetime.now(), index=[1, 22.2, 0.70, 'bob']):
        try:
            test_grid.add_period(datetime.datetime.now(), [1, 22.2, 0.70, 'bob'])
            self.assertTrue(True)
        except:
            self.assertTrue(False)

    def test_add_period_with_datetime_input(self):
        self.test_grid_add_period(period=datetime.datetime.now())

    def test_add_period_with_date_input(self):
        self.test_grid_add_period(period=datetime.date.today())

    def test_add_period_with_decimal_input(self):
        self.test_grid_add_period(index=[1, decimal.Decimal(22.2), 0.70, 'bob'])

    def test_add_period_with_float_input(self):
        self.test_grid_add_period(index=[1, float(22.2), 0.70, 'bob'])



from parchments.core.value import Value


class Block:

    def __init__(self, period, value, value_type: str, decimal_places=0, actual_number=True):
        self.period = period
        self.value = value
        self.value_type = value_type
        self.decimal_places = decimal_places
        self.actual_number = actual_number
        self.data = {
            'actual_number': self.actual_number,
            'period_key': self.period.key,
            'value_amount': Value(value, value_type).as_dict(),
        }

    def as_dict(self):
        return self.data

    def compare_historical(self, historical_block):
        if self.value_type != 'string':
            self.data['growth_amount'] = Value(self.value - historical_block.value, self.value_type, 2).as_dict()
            self.data['growth_percentage'] = Value(self.data['growth_amount']['raw'] / historical_block.value, 'percentage', 4).as_dict()

    def compare_projected(self, projected_block):
        projected_block.compare_historical(self)

    def compare_over_historical(self, over_historical_block):
        if self.value_type != 'string':
            self.data['over_growth_amount'] = Value(self.value - over_historical_block.value, self.value_type, 2).as_dict()
            self.data['over_growth_percentage'] = Value(self.data['over_growth_amount']['raw'] / over_historical_block.value, 'percentage', 4).as_dict()

    def compare_over_projected(self, over_projected_block):
        over_projected_block.compare_over_historical(self)

    def calculate_growth(self, value1, value2):
        pass


from datetime import date, datetime


class Period:

    def __init__(self, date_or_datetime, iteration):
        if type(date_or_datetime) is date:
            self.datetime = datetime.combine(date_or_datetime, datetime.min.time())
        elif type(date_or_datetime is datetime):
            self.datetime = date_or_datetime
        else:
            raise ValueError('Invalid period. Must be a datetime.date or datetime.datetime')

        self.iteration = iteration

        if self.iteration == 'year':
            self.key = '%s0101' % self.datetime.strftime('%Y')
        if self.iteration == 'month':
            self.key = '%s01' % self.datetime.strftime('%Y%m')
        if self.iteration == 'day':
            self.key = '%s' % self.datetime.strftime('%Y%m%d')
